/*
 *         	  OMEGA IRC SECURITY SERVICES
 * 	      	    (C) 2008 Omega Dev Team
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *    $Id: access.c 1433 2008-09-18 06:35:03Z twitch $
 */


#include "stdinc.h"
#include "server.h"



/*******************************************/
/**
 * find_access() - Find the FIRST matching access
 * access list entry for the given user and return
 * the access list entry.
 *
 * @param u - Pointer to the user struct
 *            that we are checking with.
 *
 * @return Access* - return a pointer to the
 *                   access entry list that the user
 *                   that the user belongs to, or
 *                   NULL if no match is found.
 *
 */


Access* find_access(User *u)
{
	dlink_node *dl,*hdl;
	char *host;
	Access *a;
	char pattern[150];

    sprintf(pattern,"%s@%s",(u->user)? u->user : "*",u->host);
	alog(DEBUG3,"Access: Checking against pattern %s",pattern);
	DLINK_FOREACH(dl,accesslist.head)
	{
		a = dl->data;


		DLINK_FOREACH(hdl,a->hosts.head)
		{
		 	host = hdl->data;

			if (match_wild(host,pattern,0))
			{
				alog(DEBUG3,"Access %s matches for pattern %s",pattern,host);
				return a;
			}
		}
	}
	return NULL;

}


/*******************************************/
/**
 * Check a users access against the specified level.
 *
 * @param u A pointer to the user structure which
 *          we are checking against.
 * @param level The level to check for.
 *
 * @return Return true if the access match, or return
 *         0 if we fail for any other reason.
 */


int check_access(User* u, uint32 level)
{

	if (!u)
		return 0;

	if (!level)
		return 1;

	alog(DEBUG3,"Checking to see if User: %s access matches %lu (Does it: %d)",
				u->nick,level,(u->access & level));

	//user has no access
	if (!u->access)
		return 0;

	if (u->access & level)
		return 1;

	return 0;


}

/*******************************************/
/**
 * setAccess() - finds then sets or removes
 * access for a user, this will set the first
 * matching access level on a user, as provided '
 * by find_access().
 *
 *  @param u - The user struct we are working with
 *
 *  @param op - wether or not we are adding/removing
 *              access from the user struct.
 *
 *  @return int - Returns true if access has been set
 *                or false if no record was found.
 */


int setAccess(User* u, int op)
{

	Access *a;

	if (!u)
		return 0;

	if (op)
	{
		if (!(a = find_access(u)))
		{
			alog(DEBUG3,"Setting access for %s (Access: %lu)",u->nick, ACC_OPER);
			u->access = ACC_OPER;
			return 0;
		}

		alog(DEBUG3,"Setting access for %s (Access: %lu)",u->nick,(unsigned long) a->flags);
        	u->access = a->flags;
		return 1;
	}

	alog(DEBUG3,"Removing access for %s",u->nick);
	u->access &= ~u->access;
	return 1;
}

/*************************************************/
/**
 *   Destroy all entries within the access list.
 *
 *   @note This destroyes all access list entries
 *         with that said, it shouldnt be called
 *         in any other circumstances besides rehash
 *         and shutdown, as it leaves us with no entries.
 *
 *
 */


int DestAccList()
{
	dlink_node *dl;
	dlink_node *tdl;

	dlink_node *hdl,*thdl; //host dlink list
	char *host;

	Access *a;

	DLINK_FOREACH_SAFE(dl,tdl,accesslist.head)
	{

		a = dl->data;

		DLINK_FOREACH_SAFE(hdl,thdl,a->hosts.head)
		{
			host = hdl->data;

			dlink_find_delete(host,&a->hosts);
			dlink_free(hdl);
			free(host);
		}


		dlink_find_delete(a,&accesslist);
		dlink_free(dl);
		free(a);

	}
	return 0;

}

/*************************************************/
/**
 *   refresh_access() - Crawls our user list and checks
 *   access on each entries host.
 *
 *   @note This method can be slower on larger nets and
 *   is designed only to update access on special instances
 *   such as rehash.
 */


void refresh_access(void)
{
    User* u = NULL;
    Access* a = NULL;

    dlink_node *udl = NULL;


     DLINK_FOREACH(udl,userlist.head)
     {
                u = udl->data;

		/**
		 * if there not an OPER or they are a SERVICE
         * they have no access. so skip that entry :)
         * this should give us back some speed and
         * proformance here. Esp on networks with more
         * Services and Users then IRCops (Which should
         * be all of them imo)
         * - Twitch
		 */

		if ((!u->oper) || (u->service))
			continue;

		setAccess(u,1);
     }

    return;
}

/* EOF */

